/***************************************************************************
 *   Copyright (C) 2006 by Gerd v. Egidy                                   *
 *   gve@intra2net.com                                                     *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string>
#include <sstream>
#include <stdexcept>
#include <iostream>

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/serialization/serialization.hpp>

#include <boost/bind.hpp>

#include "command_server.hxx"
#include "container.hxx"
#include "log.hxx"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

using namespace std;

namespace libt2n
{

command_server::command_server(server& _s)
    : s(_s)
{
    // register callback
    s.add_callback(new_connection,bind(&command_server::send_hello, boost::ref(*this), _1));
}

void command_server::send_hello(unsigned int conn_id)
{
    server_connection* sc=s.get_connection(conn_id);

    ostringstream hello;

    hello << "T2Nv" << PROTOCOL_VERSION << ';';

    int byteordercheck=1;
    hello.write((char*)&byteordercheck,sizeof(byteordercheck));

    hello << ';';

    sc->write(hello.str());
}

/// handle a command including deserialization and answering
void command_server::handle_packet(const std::string& packet, server_connection* conn)
{
    OBJLOGSTREAM(s,debug,"handling packet from connection " << conn->get_id());

    // deserialize packet
    istringstream ifs(packet);
    boost::archive::binary_iarchive ia(ifs);
    command_container ccont;

    // TODO: catch
    ia >> ccont;

    std::ostream* ostr;
    if ((ostr=s.get_logstream(fulldebug))!=NULL)
    {
        (*ostr) << "decoded packet data: " << std::endl;
        boost::archive::xml_oarchive xo(*ostr);
        xo << BOOST_SERIALIZATION_NVP(ccont);
    }

    // TODO: cast to command subclass (template)
    command *cmd=ccont.get_command();

    result_container res;

    if (cmd)
    {
        try
        {
            res.set_result((*cmd)());
        }
        catch (t2n_exception &e)
            { res.set_exception(e.clone()); }
        catch (...)
            { throw; }
    }
    else
        throw logic_error("uninitialized command called");

    ostringstream ofs;
    boost::archive::binary_oarchive oa(ofs);

    // TODO: catch
    oa << res;

    if ((ostr=s.get_logstream(fulldebug))!=NULL)
    {
        (*ostr) << "returning result, decoded data: " << std::endl;
        boost::archive::xml_oarchive xo(*ostr);
        xo << BOOST_SERIALIZATION_NVP(res);
    }

    conn->write(ofs.str());
}

/** @brief handle incoming commands
    @param[in,out] usec_timeout wait until new data is found, max timeout usecs.
            -1: wait endless, 0: instant return
*/
void command_server::handle(long long usec_timeout, long long* usec_timeout_remaining)
{
    if (s.fill_buffer(usec_timeout,usec_timeout_remaining))
    {
        string packet;
        unsigned int conn_id;

        while (s.get_packet(packet,conn_id))
            handle_packet(packet,s.get_connection(conn_id)); 
    }
    s.cleanup();
}

}
